#
# From ILUTesting, we import "ilu_threaded", "java_command", "lisp_command", "guile_command", "python_command",
#   "iluconf_dict", "makefile_dict", "iludefs_dict", and, of course, "SimpleClientServerTest"
#

from ILUTesting import *
import sys, os, string

tests = []

############################################################
#
#  First we set up a number of specific tests
#
############################################################

if iludefs_dict.has_key("ADD_PYTHON_LANGUAGE"):

	tests.append(SimpleClientServerTest("Python client / Python server, IIOP protocol",
					    "python pyClient.py", "python pyServer.py"))

	if iludefs_dict.has_key("ADD_C_LANGUAGE"):

		tests.append(SimpleClientServerTest("Python client / C server, default protocol",
						    "python pyClient.py",
						    "server"))

		if iluconf_dict.has_key("HTTP_PROTOCOL"):
			tests.append(SimpleClientServerTest("Python client / C server, HTTP 1.1 protocol",
							    "python pyClient.py",
							    "server -p http_1_1 -t tcp_0_0"))

		if iluconf_dict.has_key("IIOP_PROTOCOL"):
			tests.append(SimpleClientServerTest("Python client / C server, IIOP protocol",
							    "python pyClient.py",
							    "server -p iiop -t tcp_0_0"))

		if iluconf_dict.has_key("SUNRPC_PROTOCOL") and iluconf_dict.has_key("SUNRPCRM_TRANSPORT"):
			tests.append(SimpleClientServerTest("Python client / C server, Sun RPC protocol with sunrpcrm/tcp transport",
							    "python pyClient.py",
							    "server -p sunrpc -t sunrpcrm tcp_0_0"))

		if iluconf_dict.has_key("COURIER_PROTOCOL") and iluconf_dict.has_key("SUNRPCRM_TRANSPORT"):
			tests.append(SimpleClientServerTest("Python client / C server, Courier protocol with sunrpcrm/tcp transport",
							    "python pyClient.py",
							    "server -p courier -t sunrpcrm tcp_0_0"))

		if iluconf_dict.has_key("W3NG_PROTOCOL") and iluconf_dict.has_key("SUNRPCRM_TRANSPORT"):
			tests.append(SimpleClientServerTest("Python client / C server, w3ng protocol with sunrpcrm/tcp transports",
							    "python pyClient.py",
							    "server -p w3ng -t sunrpcrm tcp_0_0"))

	if iludefs_dict.has_key("ADD_JAVA_LANGUAGE"):

		tests.append(SimpleClientServerTest("Python client / Java server, default protocol",
						    "python pyClient.py",
						    java_command + " Test.JaServ"))

		tests.append(SimpleClientServerTest("Python client / Java 'hard' server, default protocol",
						    "python pyClient.py",
						    java_command + " Test.JaServ hard"))

	if iludefs_dict.has_key("ADD_COMMONLISP_LANGUAGE"):

		ldir = makefile_dict["ILU_LISP_DIR"]
		test = SimpleClientServerTest("Python client / Lisp server, default protocol",
					      "python pyClient.py",
					      lisp_command + ('-e', '(load "' + ldir + '/pdefsys")',
							      '-e', '(load "' + ldir + '/ilu-sysdcl")',
							      '-e', '(load "test1-sysdcl")',
							      '-e', '(pdefsys:compile-system :test1)',
							      '-e', '(pdefsys:compile-system :test1-server)',
							      '-e', '(pdefsys:load-system :test1-server)',
							      '-e', '(load "server.lisp")',
							      '-e', '(pickle-test:start-server)'));
		test.server_startup_delay = 30
		tests.append(test)

############################################################
#
#  The main() routine
#
############################################################

def run_test():

	exceptions = 0
	if len(tests) < 1:
		print 'No tests.'
	for test in tests:
		try:
			print "**************************************************"
			test.run()
		except:
			exceptions = exceptions + 1
	if (exceptions > 0):
		sys.exit(1)
	else:
		sys.exit(0)

if __name__ == "__main__":
	run_test()
